%let progname = ebv009.sas;
* xref: ebv008.sas
* input:  ebvsim.sas
* output: e009 (dataset)
* does: ALR==ORTH0 and ORTH (with optimal lambda_i)
        Efficiency for both rho and the log odds ratio
        Efficiency relative to GEE2
        True and working models for lambda_i: BB, MN, MAD
          thus there is a 3x3 layout, within each cell
          there is a range of pi and rho
        Cluster size setup is chosen by a macro variable.
********************************************************************;
libname X "";
%let n = (2 : 100)` ;
%let n = (5 : 25 )` ;
%let n = {5  25  }` ;
%let seed = 13579;
********************************************************************;
proc iml
     worksize=30000  symsize=256; * <- adjust if necessary;

%include "ebvsim.sas";
********************************************************************;
start  var2cor (v);
/*
returns corr matrix
*/

  s = 1/sqrt(vecdiag(v));
  return ( s # v # t(s) );

finish;
********************************************************************;
start get_ee_b(b0, b1, b12, pi, rho);
/*
ALR
Compute b0, b1, b12 (which define the EE)
*/
  psi= pi/(1-pi);
  a0 = 1/(1/psi+rho);  * y1+y2==0;
  a1 = -1/(1-rho);     * y1+y2==1;
  a2 = 1/(psi+rho);    * y1+y2==2;
  b0 = a0;
  b1 = a1 - a0;
  b12= a0 + a2 - 2*a1;

finish;
********************************************************************;
start get_ee(y, n, pi, rho);
/*
 y = col vec of values
 n, pi, rho = scalars
 returns the ee as a matrix, #rows = nrow(y)
*/
  g1 = (y - n#pi) / (1+rho#(n-1));
  run get_ee_b(b0, b1, b12, pi, rho);
  g2 = (0.5*b0) # n#(n-1) + b1 # (n-1)#y + (0.5*b12) # y#(y-1)  ;
  return (g1 || g2);

finish;
********************************************************************;
start var_z12 (p, r);
/*
Returns var( b0 + b1*(Y1+Y2) + b12*Y1*Y2 )
Where E[Y1] = E[Y2] = p, corr(Y1,Y2)=r
*/

  run get_ee_b(b0, b1, b12, p, r);
  q = 1 - p;
  t = p*p + r*p*q;      /*  E[Y1 Y2]  */
  return(
    2*p*q*(1+r)*b1*b1 + t*(1-t)*b12*b12 + 4*t*q*b1*b12
  );

finish;
********************************************************************;
start  derivative (d, n, p, r);
/*
ALR
- E[d g / d parms]
*/

  m = (n*n - n) /2 ;
  pq = p  * (1-p);
  c  = pq * (1-r) * (1-r) + r;
  d = j(2,2,0);
  d[1,1] = n / (1 + (n-1)*r);
  /* d[1,2] = 0; */
  d[2,1] = m * (p+p-1) * r / c;
  d[2,2] = m * pq * (1+r) / ((1-r) * c);

finish;

********************************************************************;
start  mvc  (dist, mean, var, corr, n, pi, rho);
/*
Explicit formulae
*/

  if (upcase(dist)="BB") then
    run mvc_BetaBin  (mean, var, corr, n, pi, rho);
  else if (upcase(dist)="MN") then
    run mvc_MorelNag (mean, var, corr, n, pi, rho);
  else if (upcase(dist)="MAD") then
    run mvc_Madsen   (mean, var, corr, n, pi, rho);
  else do;
    print "Unrecognized distribution" dist;
    mean=.; var=.; corr=.;
    return;
  end;

finish;
********************************************************************;
start mean_var (dist, mean, var, corr, n, pi, rho);
/*
brute force through the pdf
*/

  if (upcase(dist)="BB") then
    run betabin_d (cdf, pdf, n, pi, rho);
  else if (upcase(dist)="MN") then
    run morelnag_d (cdf, pdf, n, pi, rho);
  else if (upcase(dist)="MAD") then
    run madsen_d (cdf, pdf, n, pi, rho);
  else do;
    print "Unrecognized distribution" dist;
    mean=.; var=.; corr=.;
    return;
  end;

  g = get_ee((0:n)`, n, pi, rho);
  pdf  = pdf`;
  mean = pdf * g;       * row vec;
  g    = g - repeat(mean, nrow(g), 1);
  var  = (g` # pdf) * g;
  corr = var2cor(var)[1,2];
  mean = mean`;

finish;
********************************************************************;
start mean_var3 (mean, var, corr, n, pi, rho);
/*
E[g] and var(g) under the 3-point distributuion
*/

  g = get_ee(0 // n*pi // n, n, pi, rho);
  run threept_d (cdf, pdf, n, pi, rho);
  pdf  = pdf`;
  mean = pdf * g;       * row vec;
  g    = g - repeat(mean, nrow(g), 1);
  var  = (g` # pdf) * g;
  corr = var2cor(var)[1,2];
  mean = mean`;

finish;


********************************************************************;
start mvc_Madsen (mean, var, corr, n, p, r);
/*
ALR
Explicit formulae.
E[g], var(g), corr(g1,g2) under the Madsen model
*/

  mean = j(2,1,0);
  var  = j(2,2,0);
  phi = 1+(n-1)*r;
  pq = p*(1-p) ;
  n2 = n*n;
  m  = (n2 - n)/2;
  a  = r*(n-1)*(4*r+n*(1-r));
  b  = 2 + (4+3*n-3*n2)*r + (18-22*n+6*n2)*r*r + (n2-n) *r*r*r;
  c  = pq * (1-r) * (1-r) + r;

  num = m * pq * (a + pq * b);
  den = 2 * (1-r) * c * c;
  var[2,2] = num / den;

  num = m * (n-2) * pq * (1-p-p) * r;
  den = phi * c;
  var[1,2] = num / den;
  var[2,1] = var[1,2];

  var[1,1] = n * pq / phi;

  corr = var2cor(var)[1,2];

finish;

********************************************************************;
start mvc_BetaBin  (mean, var, corr, n, pi, rho);
/*
ALR
Explicit formulae.
E[g], var(g), corr(g1,g2) under the BetaBinomial
*/

  mean = j(2,1,0);
  var  = j(2,2,0);
  pq  = pi*(1-pi) ;
  m   = (n*n - n)/2;
  phi = 1+(n-1)*rho;

  num = m*pq*(1+(n+n-1)*rho+n*(n-1)*rho*rho);
  den = (pq*(1-rho)*(1-rho)+rho) * (1+rho-2*rho*rho);
  var[2,2] = num / den;
  var[1,1] = n * pq / phi;
  /* var[1,2] = 0; var[2,1] = 0; */

  corr = var2cor(var)[1,2];

finish;

********************************************************************;
start mvc_MorelNag (mean, var, corr, n, p, r);
/*
ALR
Explicit formulae.
E[g], var(g), corr(g1,g2) under the Morel-Nagaraj
*/

  mean = j(2,1,0);
  var  = j(2,2,0);
  pq  = p*(1-p) ;
  m   = (n*n - n)/2;
  phi = 1+(n-1)*r;

  s  = sqrt(r);
  n2 = n*n;
  a  =  -2 + (6 - 4*n)*s - (8 - 5*n + n2)*r +
        (16 - 15*n + 3*n2)*s*r   + (-10 + 11*n - 3*n2)*r*r +
        (6 - 5*n + n2)*r*r*s  ;
  b = 1 + 3*s + 2*n*r - 2*(-4 + n)*r*s +
        (-5 + 8*n - 2*n2)*r*r + (9 - 8*n + 2*n2)*r*r*s ;
  num = m * pq * (2*pq*(1-s)*(1-s)*b - r*a);
  den = (1 - p + p*r) * (r + p - p*r);
  den = 2 * den * den * (1-r)*(1+s) ;
  var[2,2] = num / den;

  num = m*(n-2)*pq*(1-p-p)*(1-s)*r*s;
  den = phi*(1+s)*(pq*(1-r)*(1-r)+r);
  var[1,2] = num / den;

  var[2,1] = var[1,2];
  var[1,1] = n * pq / phi;

  corr = var2cor(var)[1,2];

finish;

********************************************************************;
start chk_lambda(lambda, m);
  if (lambda < -1/(m-1)) then do;
    print "lambda = " lambda  " < -1/(m-1)";
  end;
  else if (lambda > 1)        then do;
    print "lambda = " lambda   " > 1";
  end;
finish;
********************************************************************;
start code_check(dummy);
/* compare brute-force compuations against explicit formulae */

  distribs = {"BB"  "MN"  "MAD"} ` ;
  do dist = 1 to nrow(distribs);
    distrib = distribs[dist];
    errors = 0;

    print
    "        n       pi        rho    c12         v11        v22   lambda";
    do pi  = 0.1 to 0.5 by 0.2;
      do n   = 2 to 40 by 2;
        do rho = 0 to 0.8 by 0.1;
          run mean_var  (distrib, mean, var, corr, n, pi, rho);
          v = vecdiag(var)` ;
          m = (n*n-n)/2;
          if (n>2) then do;
            vz12   = var_z12(pi, rho);
            lambda = (var[2,2]/(m*vz12) - 1)/(m-1);
            run chk_lambda(lambda, m);
          end;
          else lambda = .;
          * print n pi rho corr v lambda;

          run mvc  (distrib, mean1, var1, corr1, n, pi, rho);
          v1 = vecdiag(var1)` ;

          e      = sum(abs(mean1-mean)) + sum(abs(var1-var));
          norm   = sum(abs(mean1)) + sum(abs(var1));
          error  = (e > 1e-12 * norm);
          errors = errors +  error;
          if   (error) then
            print mean var, mean1 var1 ;

        end;
      end;
    end;
    print distrib errors "errors";
  end;

finish;

********************************************************************;
start get_delta (p, r);
/*
Derivative of the transformation
     (p, r) -> (p, logOR)
This derivative is used in the delta method.
*/
  d  = i(2);
  pq = p*(1-p);
  c  = pq * (1-r) * (1-r) + r;
  d[2,1] = (p+p-1)*r/(pq*c);
  d[2,2] = (1+r)/((1-r)*c);
  return (d);
finish;
********************************************************************;
start get_re (v1, v2);
/*
Ratios of diagonals
Returns a row vector (eff(pi) || eff(rho))
*/

  e = vecdiag(v2) / vecdiag(v1);

  return (e`);

finish;
********************************************************************;
start efficiency (dist_t, dist_w, nn, p, r);
/*
Efficiency of ALR==ORTH0 and of ORTH (with the
optimal lambda_i which is cluster specific)
at given true and working models for lambda_i.

nn     := col vector of cluster sizes
p, r   := parms, p=pi, r=rho
dist_t := true distrib name
dist_w := working  distrib name
*/

  ef_opt  = j(2,2,0);
  sum_d   = j(2,2,0);
  sum_hd  = j(2,2,0);
  sum_v   = j(2,2,0);
  sum_hvh = j(2,2,0);
  h = {1, 1};   * if g is the ef for ALR, then H*g is the ef for ORTH;
  * H is a diagonal matrix represented by a col vector;
  do i = 1 to nrow(nn);
    n = nn[i];
    run  derivative (d, n, p, r);                 * -E[d g / d parms];
    run mvc  (dist_t, mean, v, c, n, p, r);       * v <- true var(g);
    ef_opt  = ef_opt + d` * inv(v) * d;
    sum_d   = sum_d + d;
    sum_v   = sum_v + v;
    run mvc  (dist_w, mean, v_w, c, n, p, r);     * v_w <- working var(g);
    h[2]    = d[2,2] / v_w[2,2];
    sum_hd  = sum_hd  + h # d;
    sum_hvh = sum_hvh + h # v # h`;
  end;

  ef_alr = sum_d`  * inv(sum_v)   * sum_d;
  ef_ort = sum_hd` * inv(sum_hvh) * sum_hd;
  av_opt_rho = inv(ef_opt);
  av_alr_rho = inv(ef_alr);
  av_ort_rho = inv(ef_ort);

  /* delta method */
  delta      = get_delta (p, r); /* derivative of (p,r)->(p,logOR) */
  av_opt_lor = delta * av_opt_rho * delta`;
  av_alr_lor = delta * av_alr_rho * delta`;
  av_ort_lor = delta * av_ort_rho * delta`;

  ae_alr_rho = get_re (av_alr_rho, av_opt_rho);
  ae_alr_lor = get_re (av_alr_lor, av_opt_lor);
  ae_ort_rho = get_re (av_ort_rho, av_opt_rho);
  ae_ort_lor = get_re (av_ort_lor, av_opt_lor);
  return  (ae_alr_rho || ae_alr_lor[2]  ||
           ae_ort_rho || ae_ort_lor[2]
  ); /* order: pi rho lor; ALR, ORTH */

finish;

********************************************************************;
/* iml execution starts here -> */

*run code_check(0);
*endsas;

*reset noname;


datalbl  = {
  dist_t    dist_w     pi    rho
  ae_alr_pi ae_alr_rho ae_alr_lor
  ae_ort_pi ae_ort_rho ae_ort_lor
};
datasave = j(1, ncol(datalbl), 0);
create  X.E009  from datasave [colname=datalbl];
setout  X.E009;     * output data set;

n  = &n;
n  = shape(n, nrow(n)*ncol(n), 1); /* make sure it is a column */
print n ;

rho_vals = (0:100)/1000 || (11:99)/100;
n_rho_vals = nrow(rho_vals)*ncol(rho_vals);

distribs = {"BB"  "MN"  "MAD"} ` ;

do dist_t = 1 to nrow(distribs);
  distrib_t = distribs[dist_t];
  do dist_w = 1 to nrow(distribs);
    distrib_w = distribs[dist_w];
    do pi  = 0.1 to 0.5 by 0.1 ;
        do rho_id = 1 to n_rho_vals;
          rho  = rho_vals[rho_id];
          rho  = round(rho, 1e-3);
          pi   = round(pi , 1e-2);
          ae   = efficiency (distrib_t, distrib_w, n, pi, rho);
          datasave = dist_t || dist_w || pi || rho || ae;
          append from  datasave;
         *print distrib pi rho ae_alr;
        end;
    end;
  end;
end;

close   X.E009;     * output data set;
********************************************************************;
