%let progname = ebv011.sas;
* xref:
* input:  ebvsim.sas
* output: e011
* does: GEE2D
        Efficiency for both rho and the log odds ratio
        Efficiency relative to GEE2
        Models for lambda_i: BB, MN, MAD
        A range of pi and rho
        Cluster size setup is chosen by a macro variable.
********************************************************************;
libname X "";
%let n = (2 : 100)` ;
%let n = (5 : 25 )` ;
%let n = {5  25  }` ;
%let seed = 13579;
********************************************************************;
proc iml
     worksize=30000  symsize=512; * <- adjust if necessary;

%include "ebvsim.sas";
********************************************************************;
start  var2cor (v);
/*
returns corr matrix
*/

  s = 1/sqrt(vecdiag(v));
  return ( s # v # t(s) );

finish;
********************************************************************;
start get_ee_b(b0, b1, b12, pi, rho);
/*
GEE2D
Compute b0, b1, b12 (which define the EE)
*/
  t = pi * pi + rho * pi * (1-pi);    * E[Y1 Y2];
  b0 = -t;
  b1 = 0;
  b12= 1;

finish;
********************************************************************;
start get_ee(y, n, pi, rho);
/*
 y = col vec of values
 n, pi, rho = scalars
 returns the ee as a matrix, #rows = nrow(y)
*/
  g1 = (y - n#pi) / (1+rho#(n-1));
  run get_ee_b(b0, b1, b12, pi, rho);
  g2 = (0.5*b0) # n#(n-1) + b1 # (n-1)#y + (0.5*b12) # y#(y-1)  ;
  return (g1 || g2);

finish;
********************************************************************;
start var_z12 (p, r);
/*
Returns var( b0 + b1*(Y1+Y2) + b12*Y1*Y2 )
Where E[Y1] = E[Y2] = p, corr(Y1,Y2)=r
*/

  run get_ee_b(b0, b1, b12, p, r);
  q = 1 - p;
  t = p*p + r*p*q;      * = E[Y1 Y2];
  return(
    2*p*q*(1+r)*b1*b1 + t*(1-t)*b12*b12 + 4*t*q*b1*b12
  );

finish;
********************************************************************;
start  derivative (d, n, p, r);
/*
GEE2D
- E[d g / d parms]
*/

  d = j(2, 2, 0);
  d[1,1] = n / (1 + (n-1) * r) ;
  /* d[1,2] = 0 ; */
  m = (n*n-n)/2;
  d[2,1] =  -m * (2 * p * (1 - r) + r) ;
  d[2,2] =  -m * p * (1-p);

finish;

********************************************************************;
start  mvc  (dist, mean, var, corr, n, pi, rho);
/*
Explicit formulae
*/

  if (upcase(dist)="BB") then
    run mvc_BetaBin  (mean, var, corr, n, pi, rho);
  else if (upcase(dist)="MN") then
    run mvc_MorelNag (mean, var, corr, n, pi, rho);
  else if (upcase(dist)="MAD") then
    run mvc_Madsen   (mean, var, corr, n, pi, rho);
  else do;
    print "Unrecognized distribution" dist;
    mean=.; var=.; corr=.;
    return;
  end;

finish;
********************************************************************;
start mean_var (dist, mean, var, corr, n, pi, rho);
/*
brute force through the pdf
*/

  if (upcase(dist)="BB") then
    run betabin_d (cdf, pdf, n, pi, rho);
  else if (upcase(dist)="MN") then
    run morelnag_d (cdf, pdf, n, pi, rho);
  else if (upcase(dist)="MAD") then
    run madsen_d (cdf, pdf, n, pi, rho);
  else do;
    print "Unrecognized distribution" dist;
    mean=.; var=.; corr=.;
    return;
  end;

  g = get_ee((0:n)`, n, pi, rho);
  pdf  = pdf`;
  mean = pdf * g;       * row vec;
  g    = g - repeat(mean, nrow(g), 1);
  var  = (g` # pdf) * g;
  corr = var2cor(var)[1,2];
  mean = mean`;

finish;
********************************************************************;
start mean_var3 (mean, var, corr, n, pi, rho);
/*
E[g] and var(g) under the 3-point distributuion
*/

  g = get_ee(0 // n*pi // n, n, pi, rho);
  run threept_d (cdf, pdf, n, pi, rho);
  pdf  = pdf`;
  mean = pdf * g;       * row vec;
  g    = g - repeat(mean, nrow(g), 1);
  var  = (g` # pdf) * g;
  corr = var2cor(var)[1,2];
  mean = mean`;

finish;


********************************************************************;
start mvc_BetaBin  (mean, v, corr, n, p, r);
/* E[g], var(g), corr(g1,g2) under the BetaBinomial  */
/* GEE2D */

  mean = j(2,1,0);
  v    = j(2,2,0);

  v[1,1] = n * p * (1 - p) / (1 + (n-1)*r) ;

v[1,2]= ((-1 + n)*n*(-1 + p)*p*(p*(-1 + r) - r))/(1 + r) ;

v[2,2]= -(((-1 + n)*n*(-1 + p)*p*(p*(-1 + r) - r)*(1 + (-1 + n)*r)* (-1
+ (4 - 3*n)*r + p*(-1 + r)*(-3 + n*(2 + r))))/(2 + 6*r + 4*r**2)) ;

  v[2,1] = v[1,2];
  corr = var2cor(v)[1,2];

finish;

********************************************************************;
start mvc_MorelNag (mean, v, corr, n, p, r);
/* E[g], var(g), corr(g1,g2) under the Morel-Nagaraj */
/* Pseudo-likelihood */

  mean = j(2,1,0);
  v    = j(2,2,0);
  s = sqrt(r);

  v[1,1] = n * p * (1 - p) / (1 + (n-1)*r) ;

v[1,2]= ((-1 + n)*n*(-1 + p)*p*(-((2 + (-2 + n)*s)*r) + 2*p*(-1 + s)*(1
+ s + (-2 + n)*r)))/(2 + 2*(-1 + n)*r) ;


v[2,2]= -((-1 + n)*n*(-1 + p)*p*(2*p**2*(-1 + s)**2* (-3 + n*(2 + 4*s -
8*r) - 6*s + 9*r + 2*n**2*r) + r*(2 + 4*(-2 + n)*s + (6 - 5*n + n**2)*r)
- 2*p*(-1 + s)*(1 + s + (-11 + 6*n)*r + (9 - 8*n + 2*n**2)*r**1.5)))/4;


  v[2,1] = v[1,2];
  corr = var2cor(v)[1,2];

finish;

********************************************************************;
start mvc_Madsen (mean, v, corr, n, p, r);
/* E[g], var(g), corr(g1,g2) under the Madsen model */
/* Pseudo-likelihood */

  mean = j(2,1,0);
  v    = j(2,2,0);

  v[1,1] = n * p * (1 - p) / (1 + (n-1)*r) ;


v[1,2]= -(((-1 + n)*n*(-1 + p)*p*(-2*p*(-1 + r) + n*r))/(2 + 2*(-1 +
n)*r)) ;


v[2,2]= -((-1 + n)*n*(-1 + p)*p*((-1 + n)*n*r - p*(-1 + r)*(2 - n*r +
n**2*r) + p**2*(-1 + r)*(6 + n**2*r - n*(4 + r))))/4;

  v[2,1] = v[1,2];
  corr = var2cor(v)[1,2];

finish;

********************************************************************;
start chk_lambda(lambda, m);
  if (lambda < -1/(m-1)) then do;
    print "lambda = " lambda  " < -1/(m-1)";
  end;
  else if (lambda > 1)        then do;
    print "lambda = " lambda   " > 1";
  end;
finish;
********************************************************************;
start code_check(dummy);

  distribs = {"BB"  "MN"  "MAD"} ` ;
  do dist = 1 to nrow(distribs);
    distrib = distribs[dist];
    errors = 0;

    print
    "        n       pi        rho    c12         v11        v22   lambda";
    do pi  = 0.1 to 0.5 by 0.2;
      do n   = 2 to 40 by 2;
        do rho = 0 to 0.8 by 0.1;
          run mean_var  (distrib, mean, var, corr, n, pi, rho);
          v = vecdiag(var)` ;
          m = (n*n-n)/2;
          if (n>2) then do;
            vz12   = var_z12(pi, rho);
            lambda = (var[2,2]/(m*vz12) - 1)/(m-1);
            run chk_lambda(lambda, m);
          end;
          else lambda = .;
          * print n pi rho corr v lambda;

          run mvc  (distrib, mean1, var1, corr1, n, pi, rho);
          v1 = vecdiag(var1)` ;

          e      = sum(abs(mean1-mean)) + sum(abs(var1-var));
          norm   = sum(abs(mean1)) + sum(abs(var1));
          error  = (e > 1e-12 * norm);
          errors = errors +  error;
          if   (error) then
            print mean var, mean1 var1 ;

        end;
      end;
    end;
    print distrib errors "errors";
  end;

finish;

********************************************************************;
start get_delta (p, r);  /* derivative of (p,r)->(p,logOR) */
  d  = i(2);
  pq = p * (1-p);
  c  = pq * (1-r) * (1-r) + r;
  d[2,1] = (p+p-1)*r/(pq*c);
  d[2,2] = (1+r)/(1-r)/c;
  return (d);
finish;
********************************************************************;
start get_re (v1, v2);
/*
Ratios of diagonals
Returns a row vector (eff(pi) || eff(rho))
*/

  e = vecdiag(v2) / vecdiag(v1);

  return (e`);

finish;
********************************************************************;
start efficiency (dist, nn, p, r);
/*
nn := vector
p, r := params
dist := distrib name
*/

  ef_opt = j(2,2,0);
  sum_d   = j(2,2,0);
  sum_v   = j(2,2,0);
  do i = 1 to nrow(nn);
    n = nn[i];
    run mvc  (dist, mean, v, c, n, p, r);
    run  derivative (d, n, p, r);
    ef_opt = ef_opt + d` * inv(v) * d;
    sum_d = sum_d + d;
    sum_v = sum_v + v;
  end;

  ef_gee2d = sum_d` * inv(sum_v) * sum_d;
  av_opt_rho = inv(ef_opt);
  av_gee2d_rho  = inv(ef_gee2d);

  /* delta method */
  delta  = get_delta (p, r); /* derivative of (p,r)->(p,logOR) */
  av_gee2d_lor = delta * av_gee2d_rho * delta`;
  av_opt_lor = delta * av_opt_rho * delta`;

  ae_gee2d_rho = get_re (av_gee2d_rho, av_opt_rho);
  ae_gee2d_lor = get_re (av_gee2d_lor, av_opt_lor);
  return  (ae_gee2d_rho || ae_gee2d_lor[2]); /* order: pi rho lor */

finish;

********************************************************************;
/* iml execution starts here -> */

*reset noname;
*run code_check(0);
*endsas;

datalbl  = {dist pi rho ae_gee2d_pi ae_gee2d_rho ae_gee2d_lor};
datasave = j(1, ncol(datalbl), 0);
create  X.E011  from datasave [colname=datalbl];
setout  X.E011;     * output data set;

n  = &n;
n  = shape(n, nrow(n)*ncol(n), 1); /* make sure it is a column */
print n ;

rho_vals = (0:100)/1000 || (11:99)/100;
n_rho_vals = nrow(rho_vals)*ncol(rho_vals);

distribs = {"BB"  "MN"  "MAD"} ` ;

do dist = 1 to nrow(distribs);
  distrib = distribs[dist];
  do pi  = 0.1 to 0.5 by 0.1;
      do rho_id = 1 to n_rho_vals;
        rho  = rho_vals[rho_id];
        rho  = round(rho, 1e-3);
        pi   = round(pi , 1e-2);
        ae   = efficiency (distrib, n, pi, rho);
        datasave = dist || pi || rho || ae;
        append from  datasave;
       *print distrib pi rho ae;
      end;
  end;
end;

close   X.E011;     * output data set;
********************************************************************;
